@@ -20,12 +20,17 @@
  * Author: Colin Walters <walters@verbum.org>
  */
 
+/* for mkdtemp */
+#define _GNU_SOURCE
+
 #include "config.h"
 
 #include "ostree.h"
 #include "otutil.h"
 
 #include <sys/types.h>
+#include <stdio.h>
+#include <stdlib.h>
 #include <attr/xattr.h>
 
 #define ALIGN_VALUE(this, boundary) \
@@ -558,7 +563,7 @@ ostree_checksum_file_from_input (GFileInfo        *file_info,
   ret = TRUE;
   ot_transfer_out_value (out_csum, &ret_csum);
  out:
-  ot_clear_checksum (&checksum);
+  g_clear_pointer (&checksum, (GDestroyNotify)g_checksum_free);
   return ret;
 }
 
@@ -724,7 +729,7 @@ ostree_set_xattrs (GFile  *f,
       value_data = g_variant_get_fixed_array (value, &value_len, 1);
       
       loop_err = lsetxattr (path, (char*)name, (char*)value_data, value_len, XATTR_REPLACE) < 0;
-      ot_clear_gvariant (&value);
+      g_clear_pointer (&value, (GDestroyNotify) g_variant_unref);
       if (loop_err)
         {
           ot_util_set_error_from_errno (error, errno);
@@ -1140,6 +1145,7 @@ ostree_create_file_from_input (GFile            *dest_file,
       if (lchown (dest_path, uid, gid) < 0)
         {
           ot_util_set_error_from_errno (error, errno);
+          g_prefix_error (error, "lchown(%u, %u) failed: ", uid, gid);
           goto out;
         }
     }
@@ -1149,6 +1155,7 @@ ostree_create_file_from_input (GFile            *dest_file,
       if (chmod (dest_path, mode) < 0)
         {
           ot_util_set_error_from_errno (error, errno);
+          g_prefix_error (error, "chmod(%u) failed: ", mode);
           goto out;
         }
     }
@@ -1304,61 +1311,35 @@ ostree_create_temp_regular_file (GFile            *dir,
 }
 
 gboolean
-ostree_create_temp_hardlink (GFile            *dir,
-                             GFile            *src,
-                             const char       *prefix,
-                             const char       *suffix,
-                             GFile           **out_file,
-                             GCancellable     *cancellable,
-                             GError          **error)
+ostree_create_temp_dir (GFile            *dir,
+                        const char       *prefix,
+                        const char       *suffix,
+                        GFile           **out_file,
+                        GCancellable     *cancellable,
+                        GError          **error)
 {
   gboolean ret = FALSE;
-  int i = 0;
-  ot_lfree char *possible_name = NULL;
-  ot_lobj GFile *possible_file = NULL;
-  GString *tmp_name = NULL;
+  ot_lfree char *template = NULL;
+  ot_lobj GFile *ret_file = NULL;
 
-  tmp_name = create_tmp_string (ot_gfile_get_path_cached (dir),
-                                prefix, suffix);
+  if (dir == NULL)
+    dir = g_file_new_for_path (g_get_tmp_dir ());
+
+  template = g_strdup_printf ("%s/%s-XXXXXX",
+                              ot_gfile_get_path_cached (dir),
+                              prefix ? prefix : "tmp");
   
-  /* 128 attempts seems reasonable... */
-  for (i = 0; i < 128; i++)
+  if (mkdtemp (template) == NULL)
     {
-      if (g_cancellable_set_error_if_cancelled (cancellable, error))
-        goto out;
-
-      g_free (possible_name);
-      possible_name = subst_xxxxxx (tmp_name->str);
-      g_clear_object (&possible_file);
-      possible_file = g_file_get_child (dir, possible_name);
-
-      if (link (ot_gfile_get_path_cached (src), ot_gfile_get_path_cached (possible_file)) < 0)
-        {
-          if (errno == EEXIST)
-            continue;
-          else
-            {
-              ot_util_set_error_from_errno (error, errno);
-              goto out;
-            }
-        }
-      else
-        {
-          break;
-        }
-    }
-  if (i >= 128)
-    {
-      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
-                   "Exhausted 128 attempts to create a temporary file");
+      ot_util_set_error_from_errno (error, errno);
       goto out;
     }
 
+  ret_file = g_file_new_for_path (template);
+
   ret = TRUE;
-  ot_transfer_out_value(out_file, &possible_file);
+  ot_transfer_out_value (out_file, &ret_file);
  out:
-  if (tmp_name)
-    g_string_free (tmp_name, TRUE);
   return ret;
 }
 
@@ -1860,6 +1841,7 @@ ostree_validate_structureof_pack_superindex (GVariant      *superindex,
   csum_v = NULL;
   bloom = NULL;
 
+  g_variant_iter_free (content_iter);
   g_variant_get_child (superindex, 3, "a(ayay)", &content_iter);
   while (g_variant_iter_loop (content_iter, "(@ay@ay)",
                               &csum_v, &bloom))
